/*
 * plugin.c
 * Copyright (C) 2010 - 2013 Peter Belkner <pbelkner@snafu.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "plugin.h"
#include <Winamp/wa_ipc.h>

// http://forums.winamp.com/showthread.php?t=76553
// http://forums.winamp.com/showthread.php?p=1425144
// http://forums.winamp.com/showthread.php?p=3045918
//#define WA_TOGGLE_VIS 40192
#if ! defined (WINAMP_FILE_QUIT) // [
#define WINAMP_FILE_QUIT 40001
#define WINAMP_OPTIONS_PREFS 40012
#define WINAMP_OPTIONS_AOT 40019
#define WINAMP_FILE_REPEAT 40022
#define WINAMP_FILE_SHUFFLE 40023
#define WINAMP_HIGH_PRIORITY 40025
#define WINAMP_FILE_PLAY 40029
#define WINAMP_OPTIONS_EQ 40036
#define WINAMP_OPTIONS_ELAPSED 40037
#define WINAMP_OPTIONS_REMAINING 40038
#define WINAMP_OPTIONS_PLEDIT 40040
#define WINAMP_HELP_ABOUT 40041
#define WINAMP_MAINMENU 40043
#define WINAMP_BUTTON1 40044
#define WINAMP_BUTTON2 40045
#define WINAMP_BUTTON3 40046
#define WINAMP_BUTTON4 40047
#define WINAMP_BUTTON5 40048
#define WINAMP_VOLUMEUP 40058
#define WINAMP_VOLUMEDOWN 40059
#define WINAMP_FFWD5S 40060
#define WINAMP_REW5S 40061
#define WINAMP_NEXT_WINDOW 40063
#define WINAMP_OPTIONS_WINDOWSHADE 40064
#define WINAMP_BUTTON1_SHIFT 40144
#define WINAMP_BUTTON2_SHIFT 40145
#define WINAMP_BUTTON3_SHIFT 40146
#define WINAMP_BUTTON4_SHIFT 40147
#define WINAMP_BUTTON5_SHIFT 40148
#define WINAMP_BUTTON1_CTRL 40154
#define WINAMP_BUTTON2_CTRL 40155
#define WINAMP_BUTTON3_CTRL 40156
#define WINAMP_BUTTON4_CTRL 40157
#define WINAMP_BUTTON5_CTRL 40158
#define WINAMP_OPTIONS_DSIZE 40165
#define IDC_SORT_FILENAME 40166
#define IDC_SORT_FILETITLE 40167
#define IDC_SORT_ENTIREFILENAME 40168
#define IDC_SELECTALL 40169
#define IDC_SELECTNONE 40170
#define IDC_SELECTINV 40171
#define IDM_EQ_LOADPRE 40172
#define IDM_EQ_LOADMP3 40173
#define IDM_EQ_LOADDEFAULT 40174
#define IDM_EQ_SAVEPRE 40175
#define IDM_EQ_SAVEMP3 40176
#define IDM_EQ_SAVEDEFAULT 40177
#define IDM_EQ_DELPRE 40178
#define IDM_EQ_DELMP3 40180
#define IDC_PLAYLIST_PLAY 40184
#define WINAMP_FILE_LOC 40185
#define WINAMP_OPTIONS_EASYMOVE 40186
#define WINAMP_FILE_DIR 40187
#define WINAMP_EDIT_ID3 40188
#define WINAMP_TOGGLE_AUTOSCROLL 40189
#define WINAMP_VISSETUP 40190
#define WINAMP_PLGSETUP 40191
#define WINAMP_VISPLUGIN 40192
#define WINAMP_JUMP 40193
#define WINAMP_JUMPFILE 40194
#define WINAMP_JUMP10FWD 40195
#define WINAMP_JUMP10BACK 40197
#endif // ]

plugin_t *plugin_init(plugin_t *p, In_Module *mod, pb_str_t *root,
    pb_str_t *home)
{
  p->mod=mod;
  p->root=root;
  p->home=home;
#if defined (VIS_DELAY) // [
  p->delay_toggle_vis.hThread=NULL;
#endif // ]

  prefs_load(&p->ps.play,&p->ps.conv,home,mod->hMainWindow);

  if (NULL==decoder_init(&p->dec,&p->ps.play,mod))
    goto error;

  return p;
error:
  plugin_cleanup(p);

  return NULL;
}

plugin_t *plugin_cleanup(plugin_t *p)
{
  decoder_cleanup(&p->dec);

  if (NULL==p->home)
    free(p->home);

  if (NULL==p->root)
    free(p->root);

  return p;
}

#if defined (VIS_DELAY) // [
static DWORD WINAPI plugin_delay_vis(LPVOID lpParam)
{
  plugin_t *p=lpParam;
  HWND hMainWindow=p->mod->hMainWindow;
  int ms=VIS_DELAY_MS_Q*p->ps.play.general.vis.delay.ms+0.5;

DWRITELN("============================================================");
DVWRITELN("delay starting visualization by %d ms",ms);
DWRITELN("============================================================");
  Sleep(ms);
  PostMessage(hMainWindow,WM_COMMAND,WINAMP_VISPLUGIN,0);

  return 0;
}
#endif // ]

plugin_t *plugin_start(plugin_t *p, const in_char *fn)
{
  HWND hMainWindow=p->mod->hMainWindow;
  int visrunning=SendMessage(hMainWindow,WM_WA_IPC,0,IPC_ISVISRUNNING);

  if (NULL==decoder_start(&p->dec,fn,576))
    goto error;

//DWRITELN("=============================================");
  if (!p->dec.is_video) {
    screen_close(&p->dec.v.s);

#if defined (VIS_DELAY) // [
    if (!p->ps.play.general.vis.enabled||visrunning) {
//DMARKLN();
      p->delay_toggle_vis.hThread=NULL;
    }
    else if (0&&(!p->ps.play.general.vis.delay.enabled)) {
//DVWRITELN("delay.enabled=%d",p->ps.play.general.vis.delay.enabled);
      p->delay_toggle_vis.hThread=NULL;
      PostMessage(hMainWindow,WM_COMMAND,WINAMP_VISPLUGIN,0);
    }
    else {
//DMARKLN();
      p->delay_toggle_vis.hThread=CreateThread(
        NULL,   // _In_opt_   LPSECURITY_ATTRIBUTES lpThreadAttributes,
        0,      // _In_       SIZE_T dwStackSize,
        plugin_delay_vis,
                // _In_       LPTHREAD_START_ROUTINE lpStartAddress,
        p,      // _In_opt_   LPVOID lpParameter,
        0,      // _In_       DWORD dwCreationFlags,
        NULL    // _Out_opt_  LPDWORD lpThreadId
      );
    }
#else // ] [
    if (0!=p->ps.play.general.vis&&!visrunning)
      PostMessage(hMainWindow,WM_COMMAND,WINAMP_VISPLUGIN,0);
#endif // ]
  }
//DWRITELN("=============================================");

  return p;
error:
  return NULL;
}

void plugin_stop(plugin_t *p, int killed)
{
#if defined (VIS_DELAY) // [
  if (p->delay_toggle_vis.hThread) {
    WaitForSingleObject(p->delay_toggle_vis.hThread,INFINITE);
    CloseHandle(p->delay_toggle_vis.hThread);
    p->delay_toggle_vis.hThread=NULL;
  }
#endif // ]
  decoder_stop(&p->dec,killed?plugin_killed:NULL,p);
}

void plugin_killed(void *data)
{
  plugin_t *p=data;
  HWND hMainWindow=p->mod->hMainWindow;
  int visrunning=SendMessage(hMainWindow,WM_WA_IPC,0,IPC_ISVISRUNNING);

  if (p->dec.is_video&&!visrunning)
    screen_close(&p->dec.v.s);
}
