/*
 * vout.h
 * Copyright (C) 2010 - 2013 Peter Belkner <pbelkner@snafu.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __VOUT_H__
#define __VOUT_H__

#include <windows.h>

#ifdef __cplusplus // [
extern "C" {
typedef class IVideoOutput vout_t;
#else // ] [
// https://www.experts-exchange.com/articles/1242/Visual-C-Object-Memory-Layout.html
typedef struct _IVideoOutputVtbl IVideoOutputVtbl;
typedef struct _IVideoOutput vout_t;
typedef struct _YV12_PLANE YV12_PLANE;
typedef struct _YV12_PLANES YV12_PLANES;

#if 1 // [
#define _WINAPI
#else // ] [
#define _WINAPI WINAPI
#endif // ]
struct _IVideoOutputVtbl {
	void *__nouse1;
	void *__nouse2;
  void (_WINAPI *destroy)(vout_t *vout);
  int (_WINAPI *open)(vout_t *vout, int w, int h, int vflip, double aspectratio,
			unsigned int fmt);
  void (_WINAPI *setcallback)(vout_t *vout, LRESULT (*msgcallback)(void *token,
			HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam), void *token);
  void (_WINAPI *close)(vout_t *vout);
  void (_WINAPI *draw)(vout_t *vout, void *frame);
  void (_WINAPI *drawSubtitle)(vout_t *vout, /*SubsItem*/ void *item);
  void (_WINAPI *showStatusMsg)(vout_t *vout, const char *text);
  int (_WINAPI *get_latency)(vout_t *vout);
  void (_WINAPI *notifyBufferState)(vout_t *vout, int bufferstate); /* 0-255*/
	INT_PTR (_WINAPI *extended)(vout_t *vout, INT_PTR param1, INT_PTR param2,
			INT_PTR param3); // Dispatchable, eat this!
};

struct _IVideoOutput {
	IVideoOutputVtbl *lpVtbl;
};

struct _YV12_PLANE {
  unsigned char* baseAddr;
  long rowBytes;
} ;

struct _YV12_PLANES {
  YV12_PLANE y;
  YV12_PLANE u;
  YV12_PLANE v;
};
#endif // ]

//#define VOUT_INIT
#if defined (VOUT_INIT) // [
void vout_init(HWND hWindow);
void vout_cleanup(HWND hWindow);
#endif // ]

vout_t *vout_get(HWND hWindow);
void vout_open(vout_t *vo, int width, int height, double aspectratio);
void vout_close(vout_t *vo);

void vout_draw(vout_t *vo,
    unsigned char* yBaseAddr, long yRowBytes,
    unsigned char* vBaseAddr, long vRowBytes,
    unsigned char* uBaseAddr, long uRowBytes);
void vout_set_caption(vout_t *vo, char *buf);
HWND vout_get_hwnd(vout_t *vo);

#ifdef __cplusplus // [
}
#endif // ]

#endif // __VOUT_H__
