/*
 * ffsox_ffi.c
 * Copyright (C) 2011, 2012 Peter Belkner <pbelkner@snafu.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifdef NDEBUG /* N.B. assert used with active statements so enable always. */
#undef NDEBUG /* Must undef above assert.h or other that might include it. */
#endif

#include <assert.h>
#include "ffsox.h"

static AVCodec *ffi_find_decoder(enum AVCodecID id)
{
#if ! defined (__XXX__)
  AVCodec *p;

  switch (id) {
    case AV_CODEC_ID_MP1:
      p=avcodec_find_decoder_by_name("mp1float");
      break;
    case AV_CODEC_ID_MP2:
      p=avcodec_find_decoder_by_name("mp2float");
      break;
    case AV_CODEC_ID_MP3:
      p=avcodec_find_decoder_by_name("mp3float");
      break;
    default:
      p=NULL;
      break;
  }

  return NULL==p?avcodec_find_decoder(id):p;
#else
  AVCodec *p=AV_CODEC_ID_MP3==id?avcodec_find_decoder_by_name("mp3float"):NULL;

  return NULL==p?avcodec_find_decoder(id):p;
#endif
}

#if /*! defined (FFSOX_AUDIO_CUST) &&*/ defined (FFSOX_AUDIO_DRC) // ] [
ffi_st_t *ffi_st_init(ffi_st_t *ist, AVStream *st, int mode, double *scale)
#else // ] [
ffi_st_t *ffi_st_init(ffi_st_t *ist, AVStream *st, int mode)
#endif // ]
{
#if /*! defined (FFSOX_AUDIO_CUST) &&*/ defined (FFSOX_AUDIO_DRC) // [
  AVDictionary *opts=NULL;
  char drc_scale[128];
#endif // ]
  AVCodecContext *dc=st->codec;

#if defined (FFSOX_DEBUG) // [
fputs("---------------------------------------------\n",stderr);
fprintf(stderr,"mode.stereo=%d [%s:%d:%s]\n",FFI_AUDIO_STEREO&mode,
    pb_basenamea(__FILE__),__LINE__,__func__);
fprintf(stderr,"mode.drc.enabled=%d [%s:%d:%s]\n",FFI_AUDIO_DRC&mode,
    pb_basenamea(__FILE__),__LINE__,__func__);
fputs("---------------------------------------------\n",stderr);
#endif // ]
  ist->st=st;
#if ! defined (FFSOX_AUDIO_CUST) // [
  ist->mode=mode;
#endif // ]
  ist->dec=NULL;
  ist->dc=NULL;
  ist->index=st->index;

  ////
  ist->decoding_needed=0;
  ist->saw_first_ts=0;
  ist->dts=AV_NOPTS_VALUE;
  ist->next_dts=AV_NOPTS_VALUE;
  ////

  if (AVMEDIA_TYPE_AUDIO==dc->codec_type) {
    if (FFI_AUDIO_STEREO&mode) {
#if 0 // [
      dc->request_channel_layout=AV_CH_LAYOUT_STEREO_DOWNMIX;
      dc->request_channels=2;
#else // ] [
      dc->request_channel_layout=AV_CH_LAYOUT_STEREO;
#endif // ]
    }
    else
      dc->request_channel_layout=AV_CH_LAYOUT_NATIVE;

    dc->request_sample_fmt=AV_SAMPLE_FMT_FLT;
  }

  if (FFI_OPEN&mode) {
    if (NULL==(ist->dec=ffi_find_decoder(dc->codec_id)))
      goto error;

#if defined (FFSOX_AUDIO_DRC) // [
    if ((mode&FFI_AUDIO_DRC)&&scale) {
      sprintf(drc_scale,"%0.1f",*scale);
#if defined (FFSOX_DEBUG) // [
fprintf(stderr,"mode.drc.cale=\"%s\" [%s:%d:%s]\n",drc_scale,
    pb_basenamea(__FILE__),__LINE__,__func__);
#endif // ]

      if (0!=av_dict_set(&opts,"drc_scale",drc_scale,0))
        goto error;
    }

    if (0!=avcodec_open2(dc,ist->dec,&opts))
      goto error;

    if (opts)
      av_dict_free(&opts);
#else // ] [
    if (0!=avcodec_open2(dc,ist->dec,NULL))
      goto error;
#endif // ]
  }

  ist->dc=dc;

  return ist;
error:
#if /*! defined (FFSOX_AUDIO_CUST) &&*/ defined (FFSOX_AUDIO_DRC) // [
  if (NULL!=opts)
    av_dict_free(&opts);
#endif // ]
  ffi_st_close(ist);

  return NULL;
}

ffi_st_t *ffi_st_close(ffi_st_t *ist)
{
  if (NULL!=ist->dec&&NULL!=ist->dc) {
    avcodec_close(ist->dc);
    ist->dc=NULL;
  }
}

//#define FFI_DUMP_META
#if defined (FFI_DUMP_META) // {
void ffi_dump_meta(AVFormatContext *ic)
{
  AVDictionaryEntry *mtag=NULL;
fprintf(stderr,">>>\n");
  while (NULL!=(mtag=av_dict_get(ic->metadata,"",mtag,
      AV_DICT_IGNORE_SUFFIX))) {
fprintf(stderr,"--- \"%s\": \"%s\"\n",mtag->key,mtag->value);
  }
fprintf(stderr,"<<<\n");
}
#endif // }

#if /*! defined (FFSOX_AUDIO_CUST) &&*/ defined (FFSOX_AUDIO_DRC) // [
static ffi_t *ffi_open(ffi_t *ffi, int mode, double *scale)
#else // ] [
static ffi_t *ffi_open(ffi_t *ffi, int mode)
#endif // ]
{
  AVFormatContext *ic=ffi->ic;
#if defined (FFSOX_VST) // {
  int ai=-1,vi=-1,i;
#else // } {
  int ai=-1,i;
#endif // }

  ffi->ts_offset=AV_NOPTS_VALUE==ic->start_time?0.0:-ic->start_time;
  ffi->nb_streams=ic->nb_streams;

  if (NULL==(ffi->st=calloc(ffi->nb_streams, sizeof *ffi->st)))
    return NULL;

  for (i=0;i<ffi->nb_streams;++i) {
    AVStream *st=ic->streams[i];

#if 0
    if (st->codec->codec_type==AVMEDIA_TYPE_VIDEO)
      vi=i;
    else if (st->codec->codec_type==AVMEDIA_TYPE_AUDIO
        &&(ai<0||2==st->codec->channels)) {

      ai=i;
    }
#endif

    switch (st->codec->codec_type) {
      case AVMEDIA_TYPE_AUDIO:
        if (ai<0||2==st->codec->channels)
          ai=i;
        break;
#if defined (FFSOX_VST) // {
      case AVMEDIA_TYPE_VIDEO:
        if (vi<0)
          vi=i;
        break;
#endif // }
      default:
        break;
    }
  }

  if (ai<0)
    return NULL;

  for (i=0;i<ffi->nb_streams;++i) {
    ffi_st_t *ist=ffi->st+i;

#if /*! defined (FFSOX_AUDIO_CUST) &&*/ defined (FFSOX_AUDIO_DRC) // ] [
    if ((i==ai||i==vi)&&NULL==ffi_st_init(ist,ic->streams[i],mode,scale))
      return NULL;
#else // ] [
    if ((i==ai||i==vi)&&NULL==ffi_st_init(ist,ic->streams[i],mode))
      return NULL;
#endif // ]

    if (i==ai)
      ffi->ast=ist;
#if defined (FFSOX_VST) // {
    else if (i==vi)
      ffi->vst=ist;
#endif // }
  }
}

#if ! defined (FFO_CLONE_STREAM_META) // {
static void ffi_clone_dict(AVDictionary **od, AVDictionary *id)
{
  AVDictionaryEntry *mtag=NULL;

  while (NULL!=(mtag=av_dict_get(id,"",mtag,AV_DICT_IGNORE_SUFFIX)))
    av_dict_set(od,mtag->key,mtag->value,AV_DICT_DONT_OVERWRITE);
}

#endif // }

#if /*! defined (FFSOX_AUDIO_CUST) &&*/ defined (FFSOX_AUDIO_DRC) // [
ffi_t *ffi_init(ffi_t *ffi, pb_str_t *fn, int mode, double *scale)
#else // ] [
ffi_t *ffi_init(ffi_t *ffi, pb_str_t *fn, int mode)
#endif // ]
{
  AVFormatContext *ic=NULL;
#if ! defined (FFO_CLONE_STREAM_META) // {
  int i;
#endif // }

  FFSOX_ZERO(ffi);

//fputs("1.1 ...\n",stderr);
  if (0!=avformat_open_input(&ic,PS_UTF8(fn),NULL,NULL))
    goto error;

  ic->flags|=AVFMT_FLAG_GENPTS;
  ffi->ic=ic;

//fputs("1.2 ...\n",stderr);
  if (avformat_find_stream_info(ic,NULL)<0)
    goto error;

//fputs("1.3 ...\n",stderr);
#if ! defined (FFO_CLONE_STREAM_META) // {
  if (FFI_STREAM_META&mode) {
    for (i=0;i<ic->nb_streams;++i)
      ffi_clone_dict(&ic->metadata,ic->streams[i]->metadata);
  }
#endif // }

//fputs("1.4 ...\n",stderr);
  if (FFI_CSV&mode)
    pb_csv2avdict(ffi->ic->filename,'\t',av_dict_set,&ffi->ic->metadata);

//fputs("1.5 ...\n",stderr);
fprintf(stderr,"(FFI_OPEN&mode)=%d [%s:%d:%s]\n",
    FFI_OPEN&mode,pb_basenamea(__FILE__),__LINE__,__func__);
#if /*! defined (FFSOX_AUDIO_CUST) &&*/ defined (FFSOX_AUDIO_DRC) // ] [
  if ((FFI_OPEN&mode)&&NULL==ffi_open(ffi,mode,scale))
    goto error;
#else // ] [
  if ((FFI_OPEN&mode)&&NULL==ffi_open(ffi,mode))
    goto error;
#endif // ]

//fputs("1.6 ...\n",stderr);
  return ffi;
error:
//fputs("1.7 ...\n",stderr);
  //fprintf(stderr, "FFmpeg: Error opening \"%s\".\n", fn);
  ffi_cleanup(ffi);

//fputs("1.8 ...\n",stderr);
  return NULL;
}

ffi_t *ffi_cleanup(ffi_t *ffi)
{
  if (NULL!=ffi->st) {
    int i;

    for (i=ffi->nb_streams-1;0<=i;--i)
      ffi_st_close(ffi->st+i);

    free(ffi->st);
    ffi->st=NULL;
    ffi->ast=NULL;
#if defined (FFSOX_VST) // {
    ffi->vst=NULL;
#endif // }
  }

  if (NULL!=ffi->ic) {
    avformat_close_input(&ffi->ic);
    ffi->ic=NULL;
  }

  return ffi;
}
